class AttributeDict(dict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        self[key] = value

    def first(self, *names):
        for name in names:
            value = self.get(name)
            if value:
                return value


env = AttributeDict()




import os
import subprocess

from tempfile import NamedTemporaryFile

from contextlib import contextmanager


class ProcessResult:
    def __init__(self, stdout, stderr, returncode):
        self.stdout = stdout.strip().decode() if stdout else ''
        self.stderr = stderr.strip().decode() if stderr else ''
        self.returncode = returncode


def sh(command, capture=False, sudo=False):
    if capture:
        return subprocess.run(command, shell=True, encoding='utf8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    command = f'{sudo_cmd(sudo)} {command}'

    print(command)
    return subprocess.run(command, shell=True)



def render_percent_template(filename, destination, context, sudo=False):
    with open(filename) as f:
        with NamedTemporaryFile('w') as output:
            output.write(f.read() % context)
            output.flush()
            sh(f'{sudo_cmd(sudo)} cp {output.name} {destination}')


def render_template(filename, destination, context, sudo=False):
    with open(filename) as f:
        with NamedTemporaryFile('w') as output:
            output.write(f.read().format_map(context))
            output.flush()
            sh(f'{sudo_cmd(sudo)} cp {output.name} {destination}')


def sudo_cmd(user=None):
    if not user:
        return ''

    if user is True:
        return 'sudo '

    return f'sudo -u {user} '


def symlink(source, rule):
    sh('sudo rm -f %s' % (rule,))
    sh('sudo ln -s %s %s' % (source, rule))


@contextmanager
def cd(path):
    old = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(old)


import os

from july3 import env
from july3.rule import Rule
from july3.util import render_percent_template, sh


class TemplateRendered(Rule):
    def __init__(self, target, source, render_fn=render_percent_template, sudo=False, dependencies=None):
        self.source = source
        self.render_fn = render_fn
        self.sudo = sudo
        super().__init__(target, dependencies=[source] + (dependencies or []))

    @staticmethod
    def command(self):
        self.render_fn(self.source, self.target, env, sudo=self.sudo)


class Directory(Rule):
    def __init__(self, target, sudo=False, dependencies=None):
        self.sudo = sudo
        super().__init__(target, dependencies=dependencies)

    @staticmethod
    def command(self):
        if not os.path.exists(self.target):
            os.makedirs(self.target)


class CommandRule(Rule):
    def __init__(self, command, sudo=False, dependencies=None):
        self.sudo = sudo
        super().__init__(command, dependencies=dependencies)

    def is_made(self):
        return sh(f'type {self.target}').returncode == 0

    def updated(self):
        return 0



import os
import shutil
import unittest

from mako.template import Template

from july3 import env
from july3.rule import NoCommandSpecified, Rule
from july3.contrib.python import PythonPackage
from july3.util import sh


class TestFileRule(unittest.TestCase):

    def setUp(self):
        env.build_path = 'test-build/'
        env.web_server_names = ['codeok.net']
        env.web_server_name = 'codeok.net'
        env.project_name = 'codeok'

    def tearDown(self):
        shutil.rmtree('test-build', ignore_errors=True)

    def test_file_depends_file(self):

        @Rule(f'test-build/{env.project_name}', dependencies=['files/nginx-site.mako', 'test-build'])
        def nginx_site_file(rule):
            with open(rule.target, 'w') as f:
                f.write(Template(filename=rule.dependencies[0]).render(**env))

        @Rule('test-build')
        def build_dir(rule):
            os.makedirs(rule.target)

        nginx_site_file.make()
        self.assertTrue(env.web_server_name in open(env.build_path + env.project_name).read())

    def test_file_depends_non_existing_file(self):
        @Rule(f'test-build/{env.project_name}.x', dependencies=['files/nginx-site.mako.x'])
        def nginx_site_file(rule):
            with open(rule.rule.filename, 'w') as f:
                f.write(Template(filename=rule.dependencies[0]).render(**env))

        self.assertRaises(NoCommandSpecified, nginx_site_file.make)


class TestNonFileRule(unittest.TestCase):

    def setUp(self):
        env.build_path = 'test-build/'
        env.web_server_names = ['codeok.net']
        env.web_server_name = 'codeok.net'
        env.project_name = 'codeok'

    def tearDown(self):
        shutil.rmtree('test-build', ignore_errors=True)
        sh('yes | pip uninstall toc')

    def test_non_file_rule(self):
        PythonPackage('toc').make()
        self.assertTrue('toc' in sh('pip show toc', capture=True).stdout)

    def test_depend_non_file_rule(self):

        os.makedirs('test-build')

        toc_install = PythonPackage('toc')

        @Rule(f'test-build/{env.project_name}', dependencies=['files/nginx-site.mako', toc_install])
        def nginx_site_file(rule):
            with open(rule.target, 'w') as f:
                f.write(Template(filename=rule.dependencies[0]).render(**env))


        nginx_site_file.make()

        self.assertTrue('toc' in sh('pip show toc', capture=True).stdout)
        self.assertTrue(env.web_server_name in open(env.build_path + env.project_name).read())



import os
import shutil
import unittest

from july3 import env
from july3.contrib.git import GitPull
from july3.rule import Rule


class TestGit(unittest.TestCase):

    def tearDown(self):
        shutil.rmtree('test-build', ignore_errors=True)

    def test_updated(self):
        env.deploy_key_file = '~/.ssh/id_rsa'

        @Rule('test-build')
        def build_dir(rule):
            os.makedirs(rule.target)

        source_update = GitPull('git@github.com:youngrok/july3.git', 'test-build/july3', dependencies=[build_dir])
        source_update.make()

        self.assertTrue(os.path.exists('test-build/july3/setup.py'))





import os

from july3 import env
from july3.rule import Rule
from july3.util import sh, sudo_cmd


class PythonPackage(Rule):

    def __init__(self, *packages, dependencies=None, pip='pip3', virtualenv=None, sudo=False):
        self.packages = packages
        self.pip = pip
        self.virtualenv = virtualenv
        self.sudo = sudo

        super().__init__(str(self), dependencies)

    def is_made(self):
        return set(self.packages).issubset(
            {package.split(' ')[0] for package in sh('{0} list --format=legacy'.format(self.pip), capture=True).stdout.splitlines()}
        )

    def updated(self):
        return 0

    def __str__(self):
        return 'pip install %s' % ' '.join(self.packages)

    @staticmethod
    def command(rule):
        packages = ' '.join(rule.packages)
        pip(f'install {packages}', pip=rule.pip, virtualenv=rule.virtualenv, sudo=rule.sudo)


def pip(command, pip='pip3', virtualenv=None, sudo=False):
    pip_cmd = pip

    if virtualenv:
        pip_cmd = virtualenv + '/bin/' + pip_cmd

    pip_cmd = sudo_cmd(sudo) + pip_cmd
    sh(f'{pip_cmd} {command}')


class PythonRequirements(Rule):
    def __init__(self, requirements, dependencies=None, pip='pip3', virtualenv=None, sudo=False):

        self.requirements = requirements
        self.pip = pip
        self.virtualenv = virtualenv
        self.sudo = sudo

        super().__init__(str(self), dependencies)

    def is_made(self):
        return False

    def updated(self):
        return 0

    def __str__(self):
        return f'pip install -r --ignore-installed {self.requirements}'

    @staticmethod
    def command(rule):
        pip(f'install -r {rule.requirements}', pip=rule.pip, virtualenv=rule.virtualenv, sudo=rule.sudo)


class Virtualenv(Rule):
    def __init__(self, target, python='python', dependencies=None):
        self.python = python

        super().__init__(target, dependencies)

    def is_made(self):
        return os.path.exists(self.target + '/bin/python')

    def updated(self):
        if self.is_made():
            return os.path.getmtime(self.target + '/bin/python')
        return 0

    @staticmethod
    def command(rule):
        sh(f'{rule.python} -m venv {env.virtualenv}')


import os

from fabric.context_managers import lcd, settings
from fabric.operations import local
from fabric.state import env

from july3 import util
from july3.contrib import mako, git


def bower():
    with settings(warn_only=True):
        if local('type bower').failed:
            local('sudo npm install -g bower')


def django(project_name):
    bower()
    git.checkout(env.project_repository, env.project_path)

    mako.render_template('%s/%s/production.py.mako' % (env.project_path, env.project_name),
                             '%s/%s/production.py' % (env.project_path, env.project_name))

    local('mkdir -p %s/logs' % env.home)
    local('mkdir -p %s/static' % env.home)

    if not os.path.exists('%s/venv' % env.project_path):
        with lcd(env.project_path):
            local('virtualenv venv')

    with lcd(env.project_path):
        local('venv/bin/pip install %s -q -r %s' % (env.get('pip_options', ''), env.pip_filename))
        local('bower install')
        local('venv/bin/python manage.py syncdb --settings=%s.production' % env.project_name)
        local('venv/bin/python manage.py migrate --settings=%s.production' % env.project_name)
        local('venv/bin/python manage.py collectstatic --noinput --settings=%s.production' % env.project_name)

    local('sudo mkdir -p /etc/uwsgi')
    mako.render_builtin_template('uwsgi-app.ini.mako', '/etc/uwsgi/%s.ini' % project_name, sudo=True)
    local('sudo reload uwsgi')

    mako.render_builtin_template('nginx-site.mako', '/etc/nginx/sites-available/%s' % env.project_name, sudo=True)
    util.symlink('/etc/nginx/sites-available/%s' % env.project_name, '/etc/nginx/sites-enabled/%s' % env.project_name)
    local('sudo service nginx reload')


def ensure_mysql(root_password):
    if not package_installed('mysql-server'):
        local("sudo debconf-set-selections <<< 'mysql-server mysql-server/root_password password %s'" % root_password, shell='bash')
        local("sudo debconf-set-selections <<< 'mysql-server mysql-server/root_password_again password %s'" % root_password, shell='bash')
        local('sudo apt-get -y install mysql')

    if not local('mysqlshow -uroot -p%s' % root_password).succeeded:
        raise Exception('Cannot connect mysql server with root user')


def uwsgi_emperor():
    local('sudo pip install uwsgi virtualenv -q')

    if not os.path.exists('/etc/init/uwsgi.conf'): # check file updatd
        local('sudo cp %s/files/uwsgi.conf /etc/init/' % os.path.dirname(__file__))
        local('sudo start uwsgi')


def ensure_nodejs():
    with settings(warn_only=True):
        if not local('grep -R node /etc/apt/sources.list /etc/apt/sources.list.d/').succeeded:
            local('sudo add-apt-repository -y ppa:chris-lea/node.js')
            local('sudo apt-get update')
            local('sudo apt-get install -y nodejs')


from fabric.context_managers import settings, lcd
from fabric.operations import local


def prepare_database(root_password, user, password, database, hosts):
    with settings(warn_only=True):
        if not local('mysqlshow -u%s -p%s %s' % (user, password, database)).succeeded:
            local('mysql -uroot -p%s -e "CREATE DATABASE \\`%s\\` DEFAULT CHARACTER SET utf8;"' % (root_password, database))

            for host in hosts:
                local('mysql -uroot -p%s -e "GRANT ALL PRIVILEGES ON %s.* TO %s@\'%s\' IDENTIFIED BY \'%s\' WITH GRANT OPTION";' % (root_password, database, user, host, password))


import os.path

from july3 import env
from july3.rule import Rule
from july3.util import sh


class GitPull(Rule):

    def __init__(self, url, path, dependencies=None):
        super().__init__('git clone %s %s' % (url, path), dependencies=dependencies)
        self.url = url
        self.path = path

    def is_made(self):
        return False

    @staticmethod
    def command(rule):
        if os.path.exists(rule.path):
            sh("cd %s; ssh-agent bash -c 'ssh-add %s; git pull'" % (rule.path, env.deploy_key_file))
        else:
            sh("ssh-agent bash -c 'ssh-add %s; git clone %s %s'" % (env.deploy_key_file, rule.url, rule.path))


from july3 import env
from july3.rule import Rule
from july3.util import sh


class PostgresUser(Rule):

    def __init__(self, user, password, password_encrypted=False, superuser=False, dependencies=None):
        self.user = user
        self.password = password
        self.password_encrypted = password_encrypted
        self.superuser = superuser
        super().__init__('postgresuser:%s' % (self.user, ), dependencies)

    def is_made(self):
        return psql("SELECT COUNT(*) FROM pg_user WHERE usename = '%s';" % self.user, '-t -A', capture=True).strip() != '0'

    def updated(self):
        return 0

    @staticmethod
    def command(rule):
        options = [
            'SUPERUSER' if rule.superuser else 'NOSUPERUSER',
            'INHERIT',
            'LOGIN',
            "%s PASSWORD '%s'" % ('ENCRYPTED' if rule.password_encrypted else 'UNENCRYPTED', rule.password)
        ]
        psql("CREATE USER %s %s;" % (rule.user, ' '.join(options)))


class PostgresDatabase(Rule):

    def __init__(self, dbname, user, password, dependencies=None):
        self.dbname = dbname
        self.user = user
        self.password = password
        super().__init__('postgresdb:%s/%s' % (self.dbname, self.user), dependencies)

    def is_made(self):
        return psql(r'\l %s' % self.dbname, '-t -A', capture=True)

    def updated(self):
        return 0

    @staticmethod
    def command(rule):

        cmd = 'createdb --owner {owner} --template {template} --encoding={encoding} --lc-ctype={locale} --lc-collate={locale} {name}'.format(
            owner=rule.user,
            name=rule.dbname,
            template=env.get('postgres_template', 'template0'),
            encoding=env.get('postgres_encoding', 'UTF8'),
            locale=env.get('postgres_locale', 'en_US.UTF-8'),
        )

        if 'psql_sudo' in env:
            cmd = 'sudo -u {0} '.format(env.psql_sudo) + cmd

        sh(cmd)


class PostgresConnection(PostgresDatabase):
    def __init__(self, dbname, user, password, password_encrypted=False, superuser=False, dependencies=None):
        self.postgres_user = PostgresUser(user, password, password_encrypted, superuser)
        if not dependencies:
            dependencies = []

        dependencies.append(self.postgres_user)

        super().__init__(dbname, user, password, dependencies)
        self.name = 'postgres:%s/%s' % (dbname, user)


def psql(query, options='', capture=False):
    command = 'psql postgres %s -c "%s"' % (options, query)

    if 'psql_sudo' in env:
        command = 'sudo -u {0} '.format(env.psql_sudo) + command

    return sh(command, capture=capture).stdout


import os
import unittest

from july3.contrib.postgres import PostgresUser, PostgresDatabase, PostgresConnection
from july3.util import sh


class TestPostgresRule(unittest.TestCase):

    def tearDown(self):
        sh('dropdb j3db')
        sh('psql postgres -c "DROP USER j3test"')

    def test_postgres_user_and_db(self):

        postgres_user = PostgresUser('j3test', 'j3pass')
        postgres_db = PostgresDatabase('j3db', 'j3test', 'j3pass', dependencies=[postgres_user])

        postgres_db.make()

        self.assertEqual('1', sh(r'psql -U j3test j3db -t -c "select 1"', capture=True).stdout.strip())


    def test_postgres_connection(self):

        postgres = PostgresConnection('j3db', 'j3test', 'j3pass')

        postgres.make()

        sh(r'psql -U j3test j3db -t -c "select 1"')
        self.assertEqual('1', sh(r'psql -U j3test j3db -t -c "select 1"', capture=True).stdout.strip())




import os


class TargetNotFound(Exception):
    pass


class RuleGraph:

    def __init__(self):
        self.rules = {}

    def register(self, name, rule):
        self.rules[name] = rule

    def find_rule(self, dependency):
        if isinstance(dependency, Rule):
            return dependency

        if dependency in self.rules:
            return self.rules[dependency]

        return Rule(dependency)

    def make(self, target_name=None):
        if not target_name:
            target_name = next(iter(self.rules.keys()))

        if target_name not in self.rules:
            target_names = '\n'.join([' - ' + k for k in self.rules.keys()])
            raise TargetNotFound(f'{target_name} not found.\nAvaiable targets:\n{target_names}')

        self.rules[target_name].make()


rules = RuleGraph()


class Rule:

    def __init__(self, target, dependencies=None, **options):
        self.target = target
        self.dependencies = dependencies if dependencies else []
        self.options = options
        rules.register(str(self.target), self)

    def make(self):
        need_update = False

        for d in self.dependencies:
            dependency_rule = rules.find_rule(d)
            dependency_rule.make()

            if dependency_rule.updated() > self.updated():
                need_update = True

        if not self.is_made():
            need_update = True

        if not need_update:
            return

        if not hasattr(self, 'command'):
            raise NoCommandSpecified('Rule:' + str(self.target))

        self.command(self)

    def __str__(self):
        return str(self.target)

    def __call__(self, command):
        self.command = command
        rules.register(command.__name__, self)
        return self

    def is_made(self):
        return os.path.exists(self.target)

    def updated(self):
        if self.is_made():
            return os.path.getmtime(self.target)

        return 0


class VirtualRule(Rule):

    def is_made(self):
        return False

    def updated(self):
        return 0


class NoCommandSpecified(Exception):
    pass




import inspect

import argh
from july3 import env
from july3.rule import rules, Rule


@argh.arg('--vars', nargs='*')
@argh.arg('--print-vars')
def make(*targets, **kwargs):
    for var in kwargs.get('vars') or []:
        k, v = var.split('=')
        env[k] = v

    if kwargs.get('print-vars'):
        print(env)

    if targets:
        for target in targets:
            rules.make(target)
    else:
        rules.make()


def main(rules_dict=None):
    if not rules_dict:
        caller_frame = inspect.stack()[1]
        rules_dict = dict(inspect.getmembers(inspect.getmodule(caller_frame[0])))

    for name, rule in rules_dict.items():
        if isinstance(rule, Rule):
            rules.register(name, rule)

    argh.dispatch_command(make)


from setuptools import setup, find_packages

setup(name='july3',
      description='Deployment automation tool',
      author='Youngrok Pak',
      author_email='pak.youngrok@gmail.com',
      keywords= 'make build deploy fabric devops',
      url='https://github.com/youngrok/july3',
      version='0.0.5',
      packages=find_packages(),
      scripts=['bin/july3'],
      package_data={'july3': ['files/*'],},
      classifiers = [
                     'Development Status :: 3 - Alpha',
                     'Topic :: System :: Installation/Setup',
                     'License :: OSI Approved :: BSD License']
      )



from july3.rule import Rule
from july3.util import sh


class DebianPackage(Rule):

    def __init__(self, packages, dependencies=None):
        self.packages = packages
        super().__init__(str(self), dependencies)

    def is_made(self):
        result = sh("dpkg-query -Wf'${db:Status-abbrev}' %s" % ' '.join(self.packages), capture=True)
        return result.returncode == 0 and all([status == 'ii' for status in result.stdout.split(' ') if status])

    def updated(self):
        return 0

    def __str__(self):
        return 'apt-get install %s' % ' '.join(self.packages)

    @staticmethod
    def command(rule):
        sh('DEBIAN_FRONTEND=noninteractive sudo apt-get install --quiet -y %s' % ' '.join(rule.packages))


class Command(Rule):

    def is_made(self):
        return sh('type {0}'.format(self.target), capture=True).returncode == 0


    def updated(self):
        return 0


from tempfile import NamedTemporaryFile
from mako.template import Template

from july3.util import sh


def render_template(source, rule, context, sudo=False):
    with NamedTemporaryFile('w') as tmp:
        tmp.file.write(Template(filename=source).render(**context))
        tmp.file.flush()
        sudo_cmd = 'sudo ' if sudo else ''
        sh('%scp %s %s' % (sudo_cmd, tmp.name, rule))
        sh('%schmod a+r %s' % (sudo_cmd, rule, ))
